CREATE OR REPLACE PACKAGE dblock
   /*
   || Adds value to DBMS_LOCK by allowing easier manipulation
   || of named locks.  Calling programs use lock names only,
   || corresponding lockhandles are automatically identified,
   || used and saved for subsequent use.
   || 
   ||
   || Author:  John Beresniewicz, Savant Corp
   ||
   || 10/26/97: added expiration_secs_IN to lockhandle
   || 10/21/97: added release
   || 10/21/97: added dump_lockhandle_tbl
   || 10/17/97: created
   ||
   || Compilation Requirements: 
   ||
   || EXECUTE on DBMS_LOCK
   || EXECUTE on DBMS_SESSION
   ||
   || Execution Requirements:
   ||
   */
AS
   /* variables to anchor other variables */
   lockname_var  VARCHAR2(128);
   lockhandle_var VARCHAR2(128);

   /* 
   || returns TRUE if a COMMIT has taken place between
   || subsequent calls to the function
   || NOTE: returns TRUE on first call in session
   */
   FUNCTION committed_TF RETURN BOOLEAN;

   /*
   || returns lockhandle for given lockname, only calls
   || DBMS_LOCK.ALLOCATE_UNIQUE if lockhandle has not been
   || previously determined
   */
   FUNCTION lockhandle
      (lockname_IN IN lockname_var%TYPE
      ,expiration_secs_IN IN INTEGER := 864000)
   RETURN lockhandle_var%TYPE;

   /*
   || returns TRUE if named lock is acquired in mode
   || specified
   */
   FUNCTION get_lock_TF
      (lockname_IN IN lockname_var%TYPE
      ,mode_IN IN INTEGER := DBMS_LOCK.x_mode
      ,timeout_IN IN INTEGER := 1
      ,release_on_commit_TF IN BOOLEAN := FALSE)
   RETURN BOOLEAN;

   /* releases named lock */
   PROCEDURE release (lockname_IN IN lockname_var%TYPE);
   
   /* print contents of lockhandle_tbl for debugging */
   PROCEDURE dump_lockhandle_tbl;
   
END dblock;
/


CREATE OR REPLACE PACKAGE BODY dblock
AS
   
   /* used by committed_TF, unique to each session */
   commit_lockname  lockname_var%TYPE := 
                        DBMS_SESSION.UNIQUE_SESSION_ID;

   /* rectype to pair handles with names */
   TYPE handle_rectype IS RECORD
      (name   lockname_var%TYPE
      ,handle lockhandle_var%TYPE
      );
      
   /* table to store lockhandles by name */
   TYPE handle_tbltype IS TABLE OF handle_rectype
      INDEX BY BINARY_INTEGER;

   lockhandle_tbl handle_tbltype;

   /* asserts truth of boolean expression */
   PROCEDURE assert(v BOOLEAN) IS
   BEGIN
      IF (NOT v OR NVL(v,TRUE) )
      THEN
         RAISE program_error;
      END IF;
   END assert;


   FUNCTION committed_TF RETURN BOOLEAN
   IS
      call_status INTEGER;
   BEGIN
      /* get unique lock, expire in one day */
      call_status := DBMS_LOCK.REQUEST
                           (lockhandle => lockhandle(commit_lockname,86400)
                           ,lockmode => DBMS_LOCK.x_mode
                           ,timeout => 0
                           ,release_on_commit => TRUE);

      RETURN (call_status = 0);
   END committed_TF;


   FUNCTION get_lock_TF
      (lockname_IN IN lockname_var%TYPE
      ,mode_IN IN INTEGER := DBMS_LOCK.x_mode
      ,timeout_IN IN INTEGER := 1
      ,release_on_commit_TF IN BOOLEAN := FALSE)
   RETURN BOOLEAN
   IS
      call_status INTEGER;
      
      /* handle for the named lock */
      temp_lockhandle lockhandle_var%TYPE := lockhandle(lockname_IN);
      
   BEGIN
      call_status := DBMS_LOCK.REQUEST
                        (lockhandle => temp_lockhandle
                        ,lockmode => mode_IN
                        ,timeout => timeout_IN
                        ,release_on_commit => release_on_commit_TF
                        );
      /*
      || if lock already owned, convert to requested mode
      */
      IF call_status = 4
      THEN
         call_status := DBMS_LOCK.CONVERT
                           (lockhandle => temp_lockhandle
                           ,lockmode => mode_IN
                           ,timeout => timeout_IN
                           );
      END IF;
      
      RETURN (call_status = 0);
   END get_lock_TF;


   FUNCTION lockhandle
      (lockname_IN IN lockname_var%TYPE
      ,expiration_secs_IN IN INTEGER := 864000)
   RETURN lockhandle_var%TYPE
   IS
      call_status INTEGER;
      temp_lockhandle  lockhandle_var%TYPE;

      temp_index  BINARY_INTEGER;
      
   BEGIN
      /*
      || if lockhandle_tbl empty must call ALLOCATE_UNIQUE
      */
      IF lockhandle_tbl.COUNT = 0
      THEN
         
         DBMS_LOCK.ALLOCATE_UNIQUE
            (lockname => lockname_IN
            ,lockhandle => temp_lockhandle
            ,expiration_secs => expiration_secs_IN);
                           
         lockhandle_tbl(1).handle := temp_lockhandle;
         lockhandle_tbl(1).name := lockname_IN;

      /*
      || check lockhandle_tbl for matching lockname
      */
      ELSE
         FOR i IN lockhandle_tbl.FIRST..lockhandle_tbl.LAST
         LOOP
            IF lockhandle_tbl(i).name = lockname_IN
            THEN
               temp_lockhandle := lockhandle_tbl(i).handle;
            END IF;
         END LOOP;
      END IF;
      
      /*
      || if temp_lockhandle still null, call ALLOCATE_UNIQUE
      || and load entry into lockhandle_tbl
      */
      IF temp_lockhandle IS NULL
      THEN
         DBMS_LOCK.ALLOCATE_UNIQUE
            (lockname => lockname_IN
            ,lockhandle => temp_lockhandle);

         /*
         || add to end of lockhandle_tbl
         */
         temp_index := lockhandle_tbl.LAST+1;
         lockhandle_tbl(temp_index).handle := temp_lockhandle;
         lockhandle_tbl(temp_index).name := lockname_IN;

      END IF;
      
      RETURN temp_lockhandle;
   END lockhandle;


   PROCEDURE release (lockname_IN IN lockname_var%TYPE)
   IS
      call_status  INTEGER;
   BEGIN
      call_status := DBMS_LOCK.RELEASE
                        (lockhandle => lockhandle(lockname_IN));
   END release;

   PROCEDURE dump_lockhandle_tbl
   IS
   BEGIN
      DBMS_OUTPUT.PUT_LINE('LOCKNAME.HANDLE');
      
      FOR i IN lockhandle_tbl.FIRST..lockhandle_tbl.LAST
      LOOP
         DBMS_OUTPUT.PUT_LINE
            (lockhandle_tbl(i).name||'.'||lockhandle_tbl(i).handle);
      END LOOP;
   END dump_lockhandle_tbl;
   
END dblock;
/

