CREATE OR REPLACE PROCEDURE demohash (counter IN INTEGER)
IS
   v_row PLS_INTEGER;
   v_name VARCHAR2(30);
   hashing_failure EXCEPTION;

   /* Define the PL/SQL table */
   TYPE string_tabtype IS TABLE OF VARCHAR2(60) INDEX BY BINARY_INTEGER;
   names string_tabtype;

   /* A function which returns the hashed value. */
   FUNCTION hashval (value IN VARCHAR2) RETURN NUMBER
   IS
   BEGIN
       RETURN DBMS_UTILITY.GET_HASH_VALUE 
          (value, 37, 1073741824);  /* POWER (2, 30) */
   END hashval;

   /* Add a name to the table, using the hash function to 
      determine the row in which the value is placed. Ah, 
      the beauty of sparse PL/SQL tables! */
   PROCEDURE addname (nm IN VARCHAR2) IS
   BEGIN
       v_row := hashval (nm);

       numstr.ins (v_row, nm);

       names (v_row) := nm;
   END;

   /* Obtain the row for a name by scanning the table. */
   FUNCTION rowbyscan (nm IN VARCHAR2) RETURN PLS_INTEGER
   IS
      v_row PLS_INTEGER := names.FIRST;
      retval PLS_INTEGER;
   BEGIN
      LOOP
         EXIT WHEN v_row IS NULL;
         IF names(v_row) = nm
         THEN
            retval := v_row;
            EXIT;
         ELSE
            v_row := names.NEXT (v_row);
         END IF;
      END LOOP;
      RETURN retval;
   END;

   /* Obtain the row for a name with numstr. */
   FUNCTION rowbynumstr (nm IN VARCHAR2) RETURN PLS_INTEGER
   IS
   BEGIN
      RETURN numstr.num (nm);
   END;

   /* Obtain the row for a name by hashing the string. */
   FUNCTION rowbyhash (nm IN VARCHAR2) RETURN PLS_INTEGER
   IS
   BEGIN
      RETURN hashval (nm);
   END;

BEGIN
   numstr.clr;

   /* Don't write code like this. Encapsulate into a procedure. 
   v_row := PLVhash.keyval ('Steven');
   names (v_row) := 'Steven';

   v_row := PLVhash.keyval ('Veva');
   names (v_row) := 'Veva';
    
   etc.
   */

   /* Load up the table with a set of strings based on the number
      of iterations requested. This allows us to easily test the
      scalability of the two algorithms. */
   FOR i IN 1 .. counter
   LOOP
       addname ('Steven' || i);
       addname ('Veva' || i);
       addname ('Eli' || i);
       addname ('Chris' || i);
   END LOOP;

   /* Verify that there were no hashing conflicts (the COUNT should
      be 4 x counter. */
   p.l ('Count in names', names.COUNT);
   IF names.COUNT != 4 * counter
   THEN
      p.l ('Hashing conflict! Test suspended...');
      RAISE hashing_failure;
   END IF;

   /* Verify that the two scans return matching values. */
   v_name := 'Eli' || TRUNC (counter/2);
   p.l ('scan',rowbyscan (v_name));
   p.l ('hash',rowbyhash (v_name));
   IF rowbyscan (v_name) != rowbyhash (v_name)
   THEN
      p.l ('Scanned row differs from hashed row. Test suspended...');
      RAISE hashing_failure;
   END IF;

   /* Time performance of retrieval via scan. */
   plvtmr.capture;
   FOR i IN 1 .. counter
   LOOP
      v_row := rowbyscan (v_name);
   END LOOP;
   plvtmr.show_elapsed ('scan');

   /* Time performance of retrieval via numstr. */
   plvtmr.capture;
   FOR i IN 1 .. counter
   LOOP
      v_row := rowbynumstr (v_name);
   END LOOP;
   plvtmr.show_elapsed ('numstr');

   /* Time performance of retrieval via hashed value. */
   plvtmr.capture;
   FOR i IN 1 .. counter
   LOOP
      v_row := rowbyhash (v_name);
   END LOOP;
   plvtmr.show_elapsed ('hash');

EXCEPTION
   WHEN hashing_failure
   THEN
      NULL;
END;
/


