CREATE OR REPLACE PACKAGE my_session
   /*
   || Extends some of the functionality of DBMS_SESSION
   || and provides access to additional session-level 
   || information.
   ||
   || Author:  John Beresniewicz, Savant Corp
   ||
   || 02/04/98: fixed bug in sid function
   || 12/22/97: exposed load_unique_id as/per Phil Pitha
   || 09/07/97: modified function SID to assert WNPS
   ||           and not call load_my_session_rec 
   || 07/27/97: created
   ||
   || Compilation Requirements: 
   ||
   || SELECT on sys.v_$session   
   || SELECT on sys.v_$sesstat
   || SELECT on sys.v_$statname
   || 
   || Execution Requirements:
   ||
   || ALTER SESSION
   */
AS

   /* same as DBMS_SESSION.UNIQUE_SESSION_ID but callable in SQL */
   FUNCTION unique_id RETURN VARCHAR2;
   PRAGMA RESTRICT_REFERENCES(unique_id, WNDS,WNPS);

   /*
   || loads unique_session_id into global variable, must be called 
   || prior to using function unique_id  
   */
   PROCEDURE load_unique_id;

   /* returns session id of current session, callable in SQL */
   FUNCTION sid RETURN NUMBER;
   PRAGMA RESTRICT_REFERENCES(sid,WNDS,WNPS);

   /* closes any open database links not in use */
   PROCEDURE close_links
      (force_with_commit_TF IN BOOLEAN DEFAULT FALSE);

   /* loads session data, should be private but needs to assert purity */
   PROCEDURE load_my_session_rec;
   PRAGMA RESTRICT_REFERENCES(load_my_session_rec,WNDS);

   /* resets package states and frees memory */
   PROCEDURE reset;

   /* returns current stat value from V$SESSTAT for this session */
   FUNCTION statval(statname_IN IN VARCHAR2) RETURN NUMBER;

   /* displays session uga and pga using DBMS_OUTPUT */
   PROCEDURE memory;

   /* turns SQL tracing on/off with tag for file identification */
   PROCEDURE set_sql_trace
      (trace_TF IN  BOOLEAN
      ,tag_IN  IN  VARCHAR2 DEFAULT USER);

END my_session;
/

CREATE OR REPLACE PACKAGE BODY my_session
AS
   /* private global to hold dbms_session.unique_session_id */
   unique_id_  VARCHAR2(40);

   /*
   || my_session_cur and my_session_rec are both declared
   || to always hold all columns of V$SESSION   
   */
   CURSOR my_session_cur   
   IS
      SELECT  *
        FROM  sys.v_$session
       WHERE  audsid = USERENV('SESSIONID');

   my_session_rec  sys.v_$session%ROWTYPE;

   /* 
   || returns unique_id_ loaded by call to load_unique_id 
   */
   FUNCTION unique_id RETURN VARCHAR2
   IS
   BEGIN   
      RETURN unique_id_;
   END unique_id;

   /* 
   || returns the session id of current session 
   */
   FUNCTION sid RETURN NUMBER
   IS
      temp_session_rec sys.v_$session%ROWTYPE;
   BEGIN
      IF my_session_rec.sid IS NULL
      THEN
         OPEN my_session_cur;
         FETCH my_session_cur INTO temp_session_rec;
         CLOSE my_session_cur;
      ELSE
         temp_session_rec := my_session_rec;
      END IF;
      RETURN temp_session_rec.sid;
   END sid;

   /*
   || loads unique_session_id into global variable, must be called 
   || prior to using function unique_id  
   */
   PROCEDURE load_unique_id
   IS
   BEGIN
      unique_id_ := DBMS_SESSION.UNIQUE_SESSION_ID;
   END load_unique_id;

   /*
   || resets all package states and frees memory 
   */
   PROCEDURE reset
   IS
   BEGIN
      DBMS_SESSION.RESET_PACKAGE;
      DBMS_SESSION.FREE_UNUSED_USER_MEMORY;
   END reset;

   /*
   || closes all open database links not in use by session, 
   || or all if forced_with_commit_TF is TRUE 
   */
   PROCEDURE close_links 
      (force_with_commit_TF IN BOOLEAN DEFAULT FALSE) 
   IS

      /* declare exceptions for ORA errors */
      dblink_in_use EXCEPTION;
      PRAGMA EXCEPTION_INIT(dblink_in_use,-2080);

      dblink_not_open EXCEPTION;
      PRAGMA EXCEPTION_INIT(dblink_not_open,-2081);

      /* cursor of all db links available to user */
      CURSOR all_links_cur
      IS 
      SELECT db_link 
        FROM all_db_links;

   BEGIN
      /* try all links and close the ones you can */
      FOR dblink_rec IN all_links_cur
      LOOP
         BEGIN
            DBMS_SESSION.CLOSE_DATABASE_LINK(dblink_rec.db_link);
         EXCEPTION
            WHEN dblink_not_open
               THEN null;
            WHEN dblink_in_use 
               THEN 
                  IF force_with_commit_TF
                  THEN
                     COMMIT;
                     DBMS_SESSION.CLOSE_DATABASE_LINK(dblink_rec.db_link);
                  END IF;
            WHEN OTHERS 
               THEN null;
         END;
      END LOOP;
   END close_links;

   /*
   || loads V$SESSION data into global record for current session 
   */ 
   PROCEDURE load_my_session_rec
   IS
   BEGIN
      OPEN my_session_cur;
      FETCH my_session_cur INTO my_session_rec;
      CLOSE my_session_cur;
   END load_my_session_rec;

   /* 
   || returns current value of a statistic from 
   || V$SESSTAT for this session 
   */
   FUNCTION statval(statname_IN IN VARCHAR2) RETURN NUMBER
   IS
      CURSOR sesstat_cur (statname VARCHAR2)
      IS
         SELECT  s.value
           FROM  sys.v_$sesstat  s
                ,sys.v_$statname n
          WHERE  s.statistic# = n.statistic#
            AND  s.sid  = my_session.sid
            AND  n.name = statname;

      return_temp  NUMBER;

   BEGIN
      OPEN sesstat_cur(statname_IN);
      FETCH sesstat_cur INTO return_temp;
      CLOSE sesstat_cur;
      RETURN return_temp;

   EXCEPTION
      WHEN OTHERS THEN
         IF sesstat_cur%ISOPEN
         THEN
            CLOSE sesstat_cur;
         END IF;
         RETURN NULL;
   END statval;

   /* 
   || displays session uga and pga using dbms_output 
   */
   PROCEDURE memory
   IS
   BEGIN
      DBMS_OUTPUT.ENABLE;
      DBMS_OUTPUT.PUT_LINE('session UGA: '||
            TO_CHAR(my_session.statval('session uga memory') ) );
      DBMS_OUTPUT.PUT_LINE('session PGA: '||
            TO_CHAR(my_session.statval('session pga memory') ) );
   END memory;

   /* 
   || turns SQL tracing on/off with tag for file identification 
   */
   PROCEDURE set_sql_trace
      (trace_TF IN  BOOLEAN
      ,tag_IN  IN  VARCHAR2 DEFAULT USER)
   IS
      cursor_id  INTEGER;
   BEGIN
      DBMS_SESSION.SET_SQL_TRACE(trace_TF);

      IF trace_TF
      THEN
         cursor_id := DBMS_SQL.OPEN_CURSOR;

         /* parse a SQL stmt with the tag in it */
         DBMS_SQL.PARSE
            (cursor_id
            ,'SELECT '''||tag_IN||''' FROM DUAL'
            ,DBMS_SQL.native);
  
         DBMS_SQL.CLOSE_CURSOR(cursor_id);
      END IF;
   END set_sql_trace;
--
END my_session;
/

select * from user_errors;

